from decimal import Decimal

from obfuskey._constants import PRIME_MULTIPLIER
from obfuskey._math import next_prime
from obfuskey._types import FloatType
from obfuskey.exceptions import NegativeValueError, UnknownKeyError


def decode(
    value: str,
    alphabet: str,
) -> int:
    """
    Decodes a given string value into an integer using the provided alphabet.

    - If the value contains characters not found in the alphabet, an UnknownKeyError is
    raised.

    :param value: the string value to decode
    :param alphabet: the alphabet used to encode the value
    :return: the decoded value as an integer
    """
    if not all(c in alphabet for c in set(value)):
        raise UnknownKeyError(
            "The value contains characters not found in the current alphabet."
        )

    if len(value) == 1:
        return alphabet.index(value)

    base = len(alphabet)
    return sum(alphabet.index(c) * base ** i for i, c in enumerate(value[::-1]))


def encode(
    value: int,
    alphabet: str,
) -> str:
    """
    Encodes a given integer value into a string using the provided alphabet.

    - If the value is negative, a NegativeValueError is raised.

    :param value: an integer value to encode
    :param alphabet: the alphabet to use
    :return: the encoded value as a string
    """
    if value < 0:
        raise NegativeValueError("The value must be greater than or equal to zero.")

    if value < len(alphabet):
        return alphabet[value]

    base = len(alphabet)
    key = []

    while value > 0:
        value, i = divmod(value, base)
        key.append(alphabet[i])

    return "".join(key[::-1])


def generate_prime(
    alphabet: str,
    key_length: int,
    prime_multiplier: FloatType = PRIME_MULTIPLIER,
) -> int:
    """
    Generates a prime number using a given alphabet, key length, and prime multiplier.
    Before finding the next available prime, the maximum possible value is multiplied
    by the prime multiplier.

    :param alphabet: the alphabet, used to determine the maximum value
    :param key_length: the key length, used to determine the maximum value
    :param prime_multiplier: Optional. The multiplier to use against the maximum value
    :return: the next available prime number
    """
    if isinstance(prime_multiplier, float):
        prime_multiplier = Decimal(prime_multiplier)

    return next_prime(int((len(alphabet) ** key_length - 1) * prime_multiplier))



from math import gcd
from typing import Tuple

from obfuskey.exceptions import MaximumValueError

try:
    from math import isqrt
except ImportError:

    def isqrt(n: int) -> int:
        """
        integer square root.
        - In number theory, the integer square root (isqrt) of a positive integer n
          is the positive integer m which is the greatest integer less than or equal
          to the square root of n.
        """

        if n < 0:
            raise ValueError("Square root is not defined for negative numbers.")

        if n < 2:
            return 2

        a = 1 << ((1 + n.bit_length()) >> 1)

        while True:
            b = (a + n // a) >> 1

            if b >= a:
                return a

            a = b


def factor(n) -> Tuple[int, int]:
    """
    Solve for s, d where n - 1 = 2^s * d

    :param n: an integer being tested for primality
    :return: s, d
    """
    s = 0
    d = n - 1

    while d % 2 == 0:
        s += 1
        d //= 2

    return s, d


def is_prime(n: int) -> int:
    """
    Determines if an integer is prime

    :param n: the integer being tested
    :return: true if the integer is prime, else false
    """
    if n == 2:
        return True

    if gcd(n, 510510) > 1:
        return n in (2, 3, 5, 7, 11, 13, 17)

    if n < 2000000:
        return trial_division(n)

    return small_strong_pseudoprime(n)


def modinv(base: int, mod: int) -> int:
    try:
        return pow(base, -1, mod)
    except ValueError:
        g, _g = base, mod
        x, _x = 1, 0

        while _g:
            q = g // _g
            g, _g = _g, (g - q * _g)
            x, _x = _x, (x - q * _x)

        if g > 1:
            raise ValueError("There is no inverse for {} mod {}".format(base, mod))

        if x < 0:
            x = x + mod

        return x


def next_prime(n) -> int:
    """
    Determines the next prime after a given integer.

    If the integer is larger than 512-bit, the gmpy2 package is used. If this package
    is not installed, a MaximumValueError is raised.

    :param n: the starting integer
    :return: the next available prime
    """
    if n.bit_length() > 512:
        try:
            from gmpy2 import next_prime

            return next_prime(n)
        except ImportError:
            raise MaximumValueError(
                "For integers larger than 512-bit, you must have gmpy2 installed."
            )

    if n < 2:
        return 2

    if n < 5:
        return [3, 5, 5][n - 2]

    gap = [
        1,
        6,
        5,
        4,
        3,
        2,
        1,
        4,
        3,
        2,
        1,
        2,
        1,
        4,
        3,
        2,
        1,
        2,
        1,
        4,
        3,
        2,
        1,
        6,
        5,
        4,
        3,
        2,
        1,
        2,
    ]
    n += 1 + (n & 1)

    if n % 3 == 0 or n % 5 == 0:
        n += gap[n % 30]

    while not is_prime(n):
        n += gap[n % 30]

    return n


def small_strong_pseudoprime(n: int) -> bool:
    """
    Checks against a composite number to identify if an integer is prime or a
    pseudoprime. This checks against bases 2, 13, 23, and 1662803.

    :param n: the integer being tested
    :return: true if the integer is thought to be prime, else false
    """
    for base in [2, 13, 23, 1662803]:
        if not strong_pseudoprime(n, base):
            return False

    return True


def strong_pseudoprime(n, base=2):
    """
    Checks against a composite number to identify if an integer in a given base is
    prime or a pseudoprime.

    :param n: the integer being tested
    :param base: the base to test against
    :return: true if the integer is thought to be prime, else false
    """
    if not n & 1:
        return False

    s, d = factor(n)
    x = pow(base, d, n)

    if x == 1:
        return True

    for i in range(s):
        if x == n - 1:
            return True

        x = pow(x, 2, n)

    return False


def trial_division(n: int) -> int:
    """
    Determines if an integer is prime by using trial division.

    :param n: the integer being tested
    :return: true if the integer is prime, else false
    """
    return all(n % i for i in range(3, isqrt(n) + 1, 2))



BASE16 = "0123456789ABCDEF"
BASE36 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
BASE52 = "0123456789BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz"
BASE56 = "23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnpqrstuvwxyz"
BASE58 = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
BASE62 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
BASE64 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz+/"
BASE94 = (
    "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
)

__all__ = (BASE16, BASE36, BASE52, BASE56, BASE58, BASE62, BASE64, BASE94)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['obfuskey']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'obfuskey',
    'version': '0.1.0',
    'description': 'A small library for creating reversible, obfuscated, identifier hashes using a given alphabet.',
    'long_description': "# ObfusKey\n\nObfusKey is a utility for generating obfuscated keys of integer values. While working\nto modernize its predecessor, [BaseHash](basehash), it was found that a lot of\nsimplifications could be made, thus ObfusKey was born.\n\nObfusKey was built solely for Python 3.6 or higher. For lower versions, you can still\nuse [BaseHash][basehash].\n\nObfusKey will generate obfuscated, reversible keys using a given alphabet and key\nlength. The maximum value it can process is `base ** key_length - 1`, where `base` is\nthe length of the provided alphabet. An optional modifier can also be provided, which is\nthen required when reversing the key into it's original value. If a modifier is not \nprovided, ObfusKey will generate the next prime integer after\n`base ** key_length - 1` along with a prime modifier. The default prime modifier will\ngenerate golden ratio primes, but this can be overwritten.\n\n## Install\n\n```text\n$ pip install obfuskey\n```\n\nIf you're building from source\n\n```text\n$ pip install .\n\nOR\n\n$ poetry install\n```\n\n## Usage\n\nTo use ObfusKey, you can use one of the available alphabets, or provide your own. You\ncan also provide your own multiplier, or leave it blank to use the built-in prime\ngenerator.\n\n```python\nfrom obfuskey import ObfusKey, alphabets\n\n\nobfuscator = ObfusKey(alphabets.BASE36, key_length=8)\n\nkey = obfuscator.to_key(1234567890)     # FWQ8H52I\nvalue = obfuscator.to_value('FWQ8H52I') # 1234567890\n```\n\nTo provide a custom multiplier, or if you to provide the prime generated from a\nprevious instance, you can pass it in with `multiplier=`. This value has to be an odd\ninteger.\n\n```python\nfrom obfuskey import ObfusKey, alphabets\n\n\nobfuscator = ObfusKey(alphabets.BASE62, multiplier=46485)\nkey = obfuscator.to_key(12345) # 0cpqVJ\n```\n\nIf you wish to generate a prime not within the golden prime set, you can overwrite the\nmultiplier with\n\n```python\nfrom obfuskey import ObfusKey, alphabets\n\n\nobfuscator = ObfusKey(alphabets.BASE62, key_length=3)\nkey = obfuscator.to_key(123) # 1O9\n\nobfuscator.set_prime_multiplier(1.75)\nkey = obfuscator.to_key(123) # Fyl\n```\n\n## Extras\n\nIf you need to obfuscate integers that are larger than 512-bit, you will need to also\nhave [gmp2][gmpy2] installed.\n\n```text\n$ pip install gmpy2\n\nOR\n\npoetry install -E gmpy2\n```\n\n[basehash]: https://github.com/bnlucas/python-basehash\n[gmpy2]: https://pypi.org/project/gmpy2/",
    'author': 'Nathan Lucas',
    'author_email': 'nathan@bnlucas.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bnlucas/obfuskey',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from obfuskey._constants import KEY_LENGTH, PRIME_MULTIPLIER
from obfuskey._math import modinv
from obfuskey._types import FloatType
from obfuskey.utils import decode, encode, generate_prime
from obfuskey.exceptions import (
    DuplicateError,
    KeyLengthError,
    MaximumValueError,
    MultiplierError,
    NegativeValueError,
    UnknownKeyError,
)


class ObfusKey:
    def __init__(
        self,
        alphabet: str,
        *,
        key_length: int = KEY_LENGTH,
        multiplier: int = None,
    ):
        """
        Constructs a new instance of a given alphabet, key length, and multiplier to
        generate obfuscated keys for a given integer, or to reverse an existing key into
        it's original value.

        :param alphabet: the alphabet to use when generating keys
        :param key_length: optional. the length of the generated keys, defaults to 6
        :param multiplier: optional. the multiplier to use, defaults to a prime number
        """
        if len(set(alphabet)) != len(alphabet):
            raise DuplicateError("The alphabet contains duplicate characters.")

        if multiplier is not None and (
            not isinstance(multiplier, int) or multiplier % 2 == 0
        ):
            raise MultiplierError("The multiplier must be an odd integer.")

        self.__alphabet = alphabet
        self.__key_length = key_length
        self.__maximum_value = len(alphabet) ** key_length - 1
        self.__multiplier = multiplier
        self.__prime_multiplier = PRIME_MULTIPLIER

    @property
    def key_length(self) -> int:
        """Returns the key length that will be generated."""
        return self.__key_length

    @property
    def maximum_value(self) -> int:
        """Returns the maximum value that the instance is allowed to obfuscate."""
        return self.__maximum_value

    @property
    def multiplier(self) -> int:
        """Returns the multiplier that is being used to obfuscate values."""

        if self.__multiplier is None:
            self.__generate_multiplier()

        return self.__multiplier

    def set_prime_multiplier(self, multiplier: FloatType) -> None:
        """
        Sets the prime multiplier to be used when generating the next prime after
        base ** length - 1.

        If this is set, the existing multiplier will be removed.

        :param multiplier: either a float or Decimal value
        """

        self.__prime_multiplier = multiplier
        self.__generate_multiplier()

    def to_key(self, value: int) -> str:
        """
        Returns an obfuscated key of the given integer value.

        - If the value is negative, a NegativeValueError is raised.
        - If the value is greater than the maximum possible value, a MaximumValueError is
        raised.

        :param value: the value to generate a key from
        :return: the obfuscated key
        """

        if value < 0:
            raise NegativeValueError("The value must be greater than or equal to zero.")

        if value > self.__maximum_value:
            raise MaximumValueError(
                f"The maximum value possible is {self.__maximum_value}"
            )

        if value == 0:
            return "".rjust(self.__key_length, self.__alphabet[0])

        if self.__multiplier is None:
            self.__generate_multiplier()

        value = value * self.__multiplier % len(self.__alphabet) ** self.__key_length
        value = encode(value, self.__alphabet)

        return value.rjust(self.__key_length, self.__alphabet[0])

    def to_value(self, key: str) -> int:
        """
        Reverses an obfuscated key back to it's integer value.

        - If the key contains characters not found in the current alphabet, an
        UnknownKeyError is raised.
        - If the key's length differs from the instance's length, a KeyLengthError is
        raised.

        :param key: the key to reverse
        :return: the original integer value
        """

        if not all(c in self.__alphabet for c in set(key)):
            raise UnknownKeyError(
                "The key contains characters not found in the current alphabet."
            )

        if len(key) != self.__key_length:
            raise KeyLengthError("They key length does not match the set length.")

        if key == "".rjust(self.__key_length, self.__alphabet[0]):
            return 0

        if self.__multiplier is None:
            self.__generate_multiplier()

        key = decode(key, self.__alphabet)
        max_p1 = self.__maximum_value + 1

        return key * modinv(self.__multiplier, max_p1) % max_p1

    def __generate_multiplier(self) -> None:
        """Generates the next prime number after the instance's maximum value"""
        self.__multiplier = generate_prime(
            self.__alphabet,
            self.__key_length,
            prime_multiplier=self.__prime_multiplier,
        )



class ObfuskeyError(Exception):
    """The base exception class of BaseHash"""

    pass


class DuplicateError(ObfuskeyError):
    """Raised when the provided alphabet contains duplicate characters"""

    pass


class KeyLengthError(ObfuskeyError):
    """Raised when there is an issue with the key length"""

    pass


class MaximumValueError(ObfuskeyError):
    """Raised when the value bring processed is too large"""

    pass


class MultiplierError(ObfuskeyError):
    """Raised when the multiplier is not an odd integer"""

    pass


class NegativeValueError(ObfuskeyError):
    """Raised when a negative value is provided"""

    pass


class UnknownKeyError(ObfuskeyError):
    """Raised when an encoded value contains characters not found in the alphabet"""

    pass
